/*
    Copyright (C) 2004, 2009 Vansteenhuyse Hubert <hvansteenhuyse@yahoo.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

/*
HISTORY:
	27 jan 2004:	version 0.1. code creation, with basic functionality: just a small simple xml parser, nothing more.
	   ...     :	now and then bug fixes and improvement.
	12 Mar 2007:	GNU/GPL published on non-gnu.org
	23 sep 2007:	version 0.2. added CDATA support
	26 apr 2008:	version 0.3.
					added support for hexadecimal and decimal references
					ignores now the validation tags
					solved bugs (unclosed tags)

	18 jan 2009:	port to ansi c
	27 jan 2018: 	version 1.2 added UTF 8 support
	28 jan 2018:	added lt_set_xxx methods to allow other encodings

	20 may 2019:	fixed bug in reference (conversion of numerical references)
*/
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "lt_parser.h"

int __lt_UTF8_BOM = 0;

#define lt_utf8_len(i) 				\
	( 	((i) < 0x80) ? 1 :				\
		(((i) & 0xE0) == 0xC0) ? 2 : 	\
		(((i) & 0xF0) == 0xE0) ? 3 : 	\
		(((i) & 0xF8) == 0xF0) ? 4 :	\
		0	)


#define ISVALID_UT8(c)				\
	(	(c) <= 0x7F || 				\
		((c) & 0xE0) == 0xC0 ||		\
		((c) & 0xF0) == 0xE0 ||		\
		((c) & 0xF8) == 0xF0 )

#define ISVALID_XMLCHAR(c)  ISVALID_UT8(c)

#define LT_LCASE(c) ((c) >= 'A' && (c) <= 'Z' ? ((c) - ('A' - 'a')) : (c))
#define IS_EOL(c) ( c == 0xA ? 1 : 0 )
#define IS_BLANK(c) ( c == 0x20 || c == 0x9 || c == 0xD || c == 0xA ? 1 : 0 )
#define ISVALID_VALUECHAR(c)( c == '<' || c == '&' ? 0 : 1)
#define IS_DIGITCHAR(c)( c >= '0' && c <= '9' ? 1 : 0)
#define ISVALID_LETTER(c)( (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')  ? 1 : 0)

static char		*LT_ERROR_MEM = 			"Out of memory";
static char		*LT_ERROR_EOF =				"Unexpected EOF";
static char		*LT_ERROR_CHAR =			"Invalid char";
static char		*LT_ERROR_VALIDATION =		"Ill-formed xml validation";

int error(LX_PARSER *lx, size_t line, char *errmsg);

//UTF8 conversion
int lt_utf8_encode(unsigned char *in, unsigned char *out){
	if(in && out){
		if((*in) <= 0x7F){
			*out = *in;
			return 1;
		} else if(((*in) & 0xE0) == 0xC0){
			out[0] = ((*in) >> 6)   | 0xC0;			in++;
			out[1] = ((*in) & 0x3F) | 0x80;
			return 2;
		} else if(((*in) & 0xF0) == 0xE0){
			out[0] = ((*in) >> 12)  | 0xE0;			in++;
			out[1] = (((*in) >> 6) & 0x3F) | 0x80;	in++;
			out[2] = ((*in) & 0x3F) | 0x80;
			return 3;
		} else if(((*in) & 0xF8) == 0xF0){
			out[0] = ((*in) >> 18) | 0xF0;			in++;
			out[1] = (((*in) >> 12) & 0x3F) | 0x80;	in++;
			out[2] = (((*in) >> 6) & 0x3F) | 0x80;	in++;
			out[3] = ((*in) & 0x3F) | 0x80;
			return 4;
		}
	}
	return 0;
}
int lt_utf8_encode_codepoint(unsigned int cp, unsigned char *out){
	if(out){
		if(cp < 0x80){
			*out = cp;
			return 1;
		} else if(cp < 0x800){
			out[0] = (cp >> 6)   | 0xC0;
			out[1] = (cp & 0x3F) | 0x80;
			return 2;
		} else if(cp < 0x100000){
			out[0] = (cp >> 12)  | 0xE0;
			out[1] = ((cp >> 6) & 0x3F) | 0x80;
			out[2] = (cp & 0x3F) | 0x80;
			return 3;
		} else {
			out[0] = (cp >> 18) | 0xF0;
			out[1] = ((cp >> 12) & 0x3F) | 0x80;
			out[2] = ((cp >> 6) & 0x3F) | 0x80;
			out[3] = (cp & 0x3F) | 0x80;
			return 4;
		}
	}
	return 0;
}

unsigned int lt_utf8_codepoint(unsigned char *in, int len){
	unsigned int cp = 0;
	if(in && len){
		if(len == 1){
			cp = *in;
		} else if(len == 2){
			cp = (*in) & 0xFF;
			in++;
			cp = ((cp << 6) & 0x7FF) + ((*in) & 0x3F);
		} else if(len == 3){
			cp = (*in) & 0xFF;
			in++;
			cp = ((cp << 12) & 0xFFFF) + ((((*in) & 0xFF) << 6) & 0xFFF);
			in++;
			cp += (*in) & 0x3F;
		} else if(len == 4){
			cp = (*in) & 0xFF;
			in++;
			cp = ((cp << 18) & 0x1FFFFF) + ((((*in) & 0xFF) << 12) & 0x3FFFF);
			in++;
			cp += (((*in) & 0xFF) << 6) & 0xFFF;
			in++;
			cp += (*in) & 0x3F;
		}
	}
	return cp;
}

//returns 0 on invalid UTF8
int lt_utf8_valid(const unsigned char *str){
	int	len = 0;
	if(str){
		//check the first char and get the length
		len = lt_utf8_len(*str);
		if(len == 0){
			return 0;
		} else if(len > 1){
			//check the encoded sequence (bytes after the first char)
			for(int i = 1; i < len; i++){
				if((str[i] & 0x80) != 0x80)
					return 0;
			}
		}
	}

	return len;
}


//returns 0 on invalid UTF8
int lt_utf8_copy(unsigned char *in, unsigned char *out){
	int i = lt_utf8_valid(in);

	if(i == 1){
		*out = *in;
	} else if(i == 2){
		*out++ = *in++;
		*out++ = *in++;
	} else if(i == 3){
		*out++ = *in++;
		*out++ = *in++;
		*out++ = *in++;
	} else if(i == 4){
		*out++ = *in++;
		*out++ = *in++;
		*out++ = *in++;
		*out++ = *in++;
	}

	return i;
}


//computes the UTF8 buffer size for a string of chars
size_t lt_buflen(unsigned char *str, size_t len){
	int		n;
	size_t 	l = 0;
	while(l < len){
		n = lt_utf8_len(*str);
		if(!n) return 0;
		l += n;
		str += n;
	}
	return l;
}


//entity references an xml parser should support (following W3C recomendations)
int reference(LX_PARSER *lx, unsigned char *ref, unsigned char *result){
    int c, p, reflen = 0, val, values[3], hcount, ipow, pw, numsystem;

    if(ref && ref[0] == '&'){
		if(ref[1] == 'a' && ref[2] == 'm' && ref[3] == 'p' &&  ref[4] == ';'){ if(result) result[0] = '&'; return 5; }
		if(ref[1] == 'a' && ref[2] == 'p' && ref[3] == 'o' && ref[4] == 's' && ref[5] == ';'){ if(result) result[0] = '\''; return 6; }
		if(ref[1] == 'g' && ref[2] == 't' && ref[3] == ';'){ if(result) result[0] = '>'; return 4; }
		if(ref[1] == 'l' && ref[2] == 't' && ref[3] == ';'){ if(result) result[0] = '<'; return 4; }
		if(ref[1] == 'q' && ref[2] == 'u' && ref[3] == 'o' && ref[4] == 't' && ref[5] == ';'){ if(result) result[0] = '\"'; return 6; }
		//special formatted references...
		if(ref[1] == '#'){
            hcount = 0;
            values[0] = 0; values[1] = 0;
            if( ref[2] == 'x'){
                //hexadecimal references, e.g.: &#xHH; where HH is a hexadecimal value between 0 and FF.
				numsystem = 16;
				reflen = 3;
                do{
                    if(hcount > 1) return 0; //too many hexadecimal digits!
                    if(IS_DIGITCHAR(ref[reflen])) val = ref[reflen] - '0';
                    else if(ref[reflen] >= 'a' && ref[reflen] <= 'f') val = ref[reflen] - 'a' + 10;
                    else if(ref[reflen] >= 'A' && ref[reflen] <= 'F') val = ref[reflen] - 'A' + 10;
                    else return 0;  //no hexadecimal value
                    values[hcount] = val;
                    reflen++;
                    hcount++;
                } while(ref[reflen] != ';');

			} else{
				//decimal references, e.g.: &#DDD; where DDD is a hexadecimal value between 0 and 255.
				numsystem = 10;
				reflen = 2;
                do{
                    if(hcount > 2) return 0; //too many decimal digits!
                    if(IS_DIGITCHAR(ref[reflen])) val = ref[reflen] - '0';
                    else return 0;  //no decimal value
                    values[hcount] = val;
                    reflen++;
                    hcount++;
                } while(ref[reflen] != ';');

			}

			//convert a numerical system to an unsigned integer
			if(!hcount) return 0; //no decimal digits where present!
			val = 0;
            ipow = hcount - 1;
            for(c = 0; c < hcount; c++){
				pw = 1; for(p = 0; p < ipow; p++) pw *= numsystem;
                val += values[c] * pw;
                ipow--;
            }

            if(result){
            	unsigned int *ui = (unsigned int*) result;
            	*ui = val;
            }

            return reflen + 1;
		}
    }
    return 0;
}

/* NAME */
// valid (first) chars of a name. returns 1 if it is else 0
int ISVALID_NAMECHAR(LX_PARSER *lx, unsigned char *c, int sw_isfirst){
    if( IS_BLANK(*c) ){
    	error(lx, lx->line, "Names cannot contain blanks");
    	return 0;
    }

    int n = lt_utf8_valid(c);
    if(n == 0){
    	error(lx, lx->line, "Ill-formed char");
    	return 0;
    }

    unsigned int cp = lt_utf8_codepoint(c, n);

    if(sw_isfirst){
    	//can't start with xml but 'xml:' and 'xmlns' are name spaces
    	if(LT_LCASE(c[0]) == 'x' && LT_LCASE(c[1]) == 'm' && LT_LCASE(c[2]) == 'l'){
    		if(	c[3] != ':'
    			&&
				c[3] != 'n' && c[4] != 's'
			){
    			error(lx, lx->line, "Name can't start with xml");
    			return 0;
    		}
    	}
    }

    if(cp >= 0x20 && cp <= 0xD7FF)
    	return n;

    if(cp >= 0xE000 && cp <= 0xFFFD)
    	return n;

    if(cp >= 0x10000 && cp <= 0x10FFFF)
    	return n;


    error(lx, lx->line, "Invalid name");
    return 0;
}

unsigned char* add_string(LX_PARSER *lx, unsigned char *dst, unsigned char *str, size_t sze, int sw_parse){
	unsigned char *p;
	size_t 	i,
			s = (size_t)( dst ? strlen((char*)dst) : 0 ),	//append to dst
			bufsze = lt_buflen(str, sze);
	int blank, refcnt, rlen;
	unsigned int esc;


	if(sze > 0 && bufsze == 0){
		error(lx, lx->line, LT_ERROR_CHAR);
		return dst;
	}

	if(!dst)	p = (unsigned char*) malloc(bufsze + 2);
	else		p = (unsigned char*) realloc(dst, 2 + s + bufsze);


	if(p){
		memset(p + s, 0, bufsze + 2);
		if(s){ p[s] = ' '; s++; }

		i = 0;
		blank = 0;
		if(sw_parse){
			while(sze){
				while(sze && IS_BLANK(str[i])){ i++; sze--; blank = 1; }
				while(sze && !IS_BLANK(str[i])){
					if(blank){p[s] = ' '; s++; blank = 0; }
					if(str[i] == '&'){
						refcnt = reference(lx, &str[i], (unsigned char*)&esc);
						i += refcnt;
						sze -= refcnt;
						s += lt_utf8_encode_codepoint(esc, &p[s]);
					} else {
						if(ISVALID_UT8(str[i])){
							//copy utf8
							refcnt = lt_utf8_copy(&str[i], &p[s]);
							if(refcnt == 0){
								error(lx, lx->line, LT_ERROR_CHAR);
								return dst;
							}
							s += refcnt;
							sze -= refcnt;
							i+=refcnt;
						} else {
							//encode utf8
							refcnt = lt_utf8_encode(&str[i], &p[s]);
							s += refcnt;
							sze -= refcnt;
							i+=refcnt;
							//i++;
						}
					}
				}
			}
		} else {	//sw_parse
			while(sze){
				//s += lt_encode_utf8(str[i], &p[s]);
				//i++; sze--;
				if(ISVALID_UT8(str[i])){
					//copy utf8
					refcnt = lt_utf8_copy(&str[i], &p[s]);
					if(refcnt == 0){
						error(lx, lx->line, LT_ERROR_CHAR);
						return dst;
					}
					s += refcnt;
					sze -= refcnt;
					i+=refcnt;
				} else {
					//encode utf8
					refcnt = lt_utf8_encode(&str[i], &p[s]);
					s += refcnt;
					sze -= refcnt;
					i+=refcnt;
					//i++;
				}

			}
		}
	}
	return p;
}

//other encodings can use these methods to assign a new buffer
void lt_set_name(XML_DATA *data, unsigned char *name){
	if(data && data->name != name){
		free(data->name);
		data->name = name;
	}
}
void lt_set_value(XML_DATA *data, unsigned char *value){
	if(data && data->val != value){
		free(data->val);
		data->val = value;
	}
}

LX_PARSER* lt_new(){
	LX_PARSER *lx = malloc(sizeof(LX_PARSER));
	memset(lx, 0, sizeof(LX_PARSER));
	return lx;
}

XML_DATA* new_data(LX_PARSER *lx, XML_DATA_TYPE type, size_t group_id){
	size_t sze;
	XML_DATA *p = 0;
		if(!lx) return 0;
		p = (XML_DATA*)malloc(sizeof(XML_DATA));
		if(!p) return 0;
		sze = (1 + lx->datacounter) * sizeof(XML_DATA*);
		p->type = type;
		p->group_id = group_id;
		p->line = lx->line;
		p->name = 0;
		p->val = 0;
		p->parent = 0;
		p->sw_tag_open = 0;
        if(!lx->datacounter) lx->data = (XML_DATA**)malloc(sze); else lx->data = (XML_DATA**)realloc(lx->data, sze);
        if(!lx->data){ free(p); return 0; }
		lx->data[lx->datacounter] = p;
        lx->datacounter++;
        return p;
}


LX_PARSER* lt_reset(LX_PARSER *lx){
	size_t i;
	XML_DATA *elem;
	if(lx && lx->data){
		for(i = 0; i < lx->datacounter; i++){
			elem = lx->data[i];
			if(elem){
				if(elem->name) free(elem->name);
				if(elem->val) free(elem->val);
				free(elem);
			}
		}
		free(lx->data);
    }

	lx->data = 0;
	lx->datacounter = 0;
	lx->line = 0;
	lx->bufsze = 0;

	return lx;
}

LX_PARSER* lt_delete(LX_PARSER *lx){
	if(lx){
		lx = lt_reset(lx);
		free(lx);
	}
	return 0;
}

//accept only plain ascii as encoding
int encoding(LX_PARSER* lx){
	unsigned short utf16;
	unsigned int utf32;

    lx->idx = lx->buf;
	if(lx->bufsze >=4){
		//get the signature for UTF 16
		utf16 = (unsigned char)lx->buf[0] << 8 | (unsigned char)lx->buf[1];

		//is it utf32?
		utf32 = utf16 << 16 | (unsigned char)lx->buf[2] << 8 | (unsigned char)lx->buf[3];
		if(utf32 == 0xFFFE0000) return error(lx, 0, "UTF32-LE not supported.");
		if(utf32 == 0x0000FFFE) return error(lx, 0, "UTF32-BE not supported.");

		//is it UTF16 (BOM required)
		if(utf16 == 0xFFFE) return error(lx, 0, "UTF16-LE not supported.");
		if(utf16 == 0xFEFF) return error(lx, 0, "UTF16-BE not supported.");

		//is it UTF8 with a BOM (optional)
		__lt_UTF8_BOM = 0;
	    if((unsigned char)lx->buf[0] == 0xEF && (unsigned char)lx->buf[1] == 0xBB && (unsigned char)lx->buf[2] == 0xBF){
	    	__lt_UTF8_BOM = 1;
	    	lx->idx += 3;
	    }

	}
	return 1; //ok, encoding is supported
}

//find next non blank char, count the lines
unsigned char* lt_next(LX_PARSER *lx){
    int n;
		if(!lx) return 0;
        while((n = (int)lx->idx[0])){
          if(!ISVALID_XMLCHAR(n)){
        	  error(lx, lx->line, LT_ERROR_CHAR);
        	  return 0;
          }
          if(IS_EOL(n)) lx->line++;
          if(!IS_BLANK(n)) return lx->idx;
          lx->idx++;
        };
        return 0;
}


//parse the buffer and build the tags sequentially in memory
int parser(LX_PARSER *lx, XML_DATA *parent){
	int refcnt;
	unsigned char *start, *tagcontent, att_quote;
	size_t start_len;
	XML_DATA *xdata = 0, *attrib = 0, *tag_close = 0, *opened_tag = 0;
	if(!lx) return 0;
	while(lt_next(lx)){
		if(lx->idx[0] == '<'){
			if(lx->idx[1] == '/'){
				//closing tag
				lx->idx+=2;
				lt_next(lx);
				opened_tag = (parent ? parent : xdata);
				start = opened_tag->name;
				if(!start) return error(lx, lx->line, "Invalid tag closure");
				if(opened_tag->type == XML_TAG) opened_tag->sw_tag_open = 0;
				start_len = 0;
				while(lx->idx[0]){
					if(lx->idx[0] == '>'){
						tag_close = new_data(lx, XML_TAG_CLOSE, (parent ? parent->group_id : xdata->group_id));
						if(!tag_close) return error(lx, lx->line, LT_ERROR_MEM);
						tag_close->parent = (parent ? parent : xdata);
						return 1;
					}
					if(lx->idx[0] != start[start_len]) return error(lx, lx->line, "Invalid tag closure");
					if(IS_EOL(lx->idx[0]))lx->line++;
					lx->idx++;
					start_len++;
				}
                return error(lx, lx->line, LT_ERROR_EOF);

			} else if(lx->idx[1] == '!' && lx->idx[2] == '-' && lx->idx[3] == '-'){
            	//<!-- ...comment... -->
            	while(lx->idx[0]){
					if(lx->idx[0] == '-' && lx->idx[1] == '-' && lx->idx[2] == '>') break;
					if(IS_EOL(lx->idx[0]))lx->line++;
					lx->idx++;
				};
            	if(lx->idx) lx->idx += 3; else  return error(lx, lx->line, LT_ERROR_EOF);
            	continue;
			}else if(lx->idx[1] == '!' && lx->idx[2] == 'D' && lx->idx[3] == 'O' && lx->idx[4] == 'C' && lx->idx[5] == 'T' && lx->idx[6] == 'Y' && lx->idx[7] == 'P' && lx->idx[8] == 'E'){
				//Skip validation tags: <!DOCTYPE name [ ... ]>
				lx->idx += 9;
				refcnt = 0;
				while(*lx->idx){
					if(IS_EOL((*lx->idx))) lx->line++;
					if(*lx->idx == '['){
						if(refcnt) return error(lx, lx->line, LT_ERROR_VALIDATION);
						refcnt = 1;
					} else if(!refcnt && *lx->idx == '>'){
						break;
					} else if(refcnt && *lx->idx == ']'){
						lx->idx++;
						if(*lx->idx != '>') return error(lx, lx->line, LT_ERROR_VALIDATION);
						break;
					}
					lx->idx++;
				}
				if(!(*lx->idx)) return error(lx, lx->line, LT_ERROR_EOF);
				lx->idx++;
				continue;
				/*
				if(!IS_BLANK(lx->idx[0])) return error(lx, lx->line, "Ill-formed validation tag");
				while(IS_BLANK(lx->idx[0])) lx->idx++;
				start = lx->idx; start_len = 0;
				while( lx->idx[0] && !IS_BLANK(lx->idx[0]) ){
					refcnt = ISVALID_NAMECHAR(lx, &lx->idx[0], (start_len == 0 ? 1 : 0));
					if(lx->sw_error) return 0;
					lx->idx += refcnt;
					if(IS_EOL(lx->idx[0])) lx->line++;
                };
				while(IS_BLANK(lx->idx[0])) lx->idx++;
				if(lx->idx[0] != '[') return error(lx, lx->line, "Ill-formed validation tag");
				lx->idx++;
				while(lx->idx[0]){ if(IS_EOL(lx->idx[0])) lx->line++; if(lx->idx[0] == ']' && lx->idx[1] == '>') break; lx->idx++; }
				if(lx->idx[0]) lx->idx += 2; else return error(lx, lx->line, LT_ERROR_EOF);
				continue;
				*/

			}else if(lx->idx[1] == '?' && lx->idx[2] == 'x' && lx->idx[3] == 'm' && lx->idx[4] == 'l'){
				//<?xml ...prolog: don't care... ?>
            	while(lx->idx[0]){
					if(lx->idx[0] == '?' && lx->idx[1] == '>') break;
					if(IS_EOL(lx->idx[0]))lx->line++;
					lx->idx++;
				};
            	if(lx->idx) lx->idx += 2; else  return error(lx, lx->line, LT_ERROR_EOF);
            	continue;
			} else if(lx->idx[1] == '!' && lx->idx[2] == '[' && lx->idx[3] == 'C' && lx->idx[4] == 'D' && lx->idx[5] == 'A' && lx->idx[6] == 'T' && lx->idx[7] == 'A' && lx->idx[8] == '['){
				//<![CDATA[ ... unparsed data... ]]>
				if(!parent || parent->type != XML_TAG) return error(lx, lx->line, "CDATA not permitted");
				start = lx->idx; start_len = 0;
				while(lx->idx[0] && !(lx->idx[1] == ']' && lx->idx[2] == ']' && lx->idx[3] == '>')){
					if(IS_EOL(lx->idx[0])) lx->line++;
					if(!lx->idx[0]) return error(lx, lx->line, LT_ERROR_EOF);
					lx->idx++; start_len++;
				}
				if(!lx->idx[0]) return error(lx, lx->line, LT_ERROR_EOF);
				tagcontent = parent->val;
				tagcontent = add_string(lx, tagcontent, start, start_len+4, 0);	//false (0): do not parse the data!!!
				parent->val = tagcontent;
				lx->idx+=4;
				continue;
			} else {
				//in a tag...
				lx->idx++; start = lx->idx; start_len = 0;
				while(lx->idx[0] && !IS_BLANK(lx->idx[0]) && lx->idx[0] != '/' && lx->idx[0] != '>'){
					refcnt = ISVALID_NAMECHAR(lx, &lx->idx[0], (start_len == 0 ? 1 : 0));
					if(lx->sw_error) return 0;
					start_len += refcnt;
					lx->idx += refcnt;
					if(IS_EOL(lx->idx[0])) lx->line++;
                }
				if(!lx->idx[0]) return error(lx, lx->line, LT_ERROR_EOF);
                xdata = new_data(lx, XML_TAG, lx->datacounter);
                if(!xdata) return error(lx, lx->line, LT_ERROR_MEM);
                xdata->parent = parent;
                xdata->name = add_string(lx, xdata->name, start, start_len, 1);
                if(lx->sw_error) return 1;
                if(!xdata->name) return error(lx, lx->line, LT_ERROR_MEM);
                //zoeken naar '>'. '/>' of een tag attributen
                while(lx->idx[0] && *lt_next(lx)){
					if(lx->idx[0] == '>') break;
					if(lx->idx[0] == '/' && lx->idx[1] == '>') break;
					//attribuut
					attrib = new_data(lx, XML_ATTRIB, xdata->group_id);
                	if(!attrib) return error(lx, lx->line, LT_ERROR_MEM);
                	attrib->parent = xdata;
                	start = lx->idx; start_len = 0;
					while(lx->idx[0] && !IS_BLANK(lx->idx[0])){
						if(lx->idx[0] == '=') break;
						refcnt = ISVALID_NAMECHAR(lx, &lx->idx[0], (start_len == 0 ? 1 : 0));
						if(lx->sw_error) return 0;
						start_len += refcnt;
						lx->idx += refcnt;
						if(IS_EOL(lx->idx[0])) lx->line++;
	                }
					if(!lx->idx[0]) return error(lx, lx->line, LT_ERROR_EOF);

	                attrib->name = add_string(lx, attrib->name, start, start_len, 1);
	                if(lx->sw_error) return 1;
	                if(!attrib->name) return error(lx, lx->line, LT_ERROR_MEM);
                    if(lx->idx[0] != '=') lt_next(lx);
                    if(lx->idx[0] != '=') return error(lx, lx->line, "Ill-formed attribute");
                    lx->idx++;
                    if(IS_BLANK(lx->idx[0])) lt_next(lx);
                    att_quote = lx->idx[0];
					if(lx->idx[0] != '\'' && lx->idx[0] != '"') return error(lx, lx->line, "Ill-formed attribute");
					lx->idx++;
					start = lx->idx; start_len = 0;
					while(lx->idx[0]){
						if(lx->idx[0] == att_quote) break;
						if( !ISVALID_VALUECHAR(lx->idx[0]) ) return error(lx, lx->line, "Invalid attribute value");
						start_len++;
						lx->idx++;
						if(IS_EOL(lx->idx[0])) lx->line++;
	                }
					if(!lx->idx[0]) return error(lx, lx->line, LT_ERROR_EOF);

					attrib->val = add_string(lx, attrib->val, start, start_len, 1);
	                if(lx->sw_error) return 1;
					if(!attrib->val) return error(lx, lx->line, LT_ERROR_MEM);
					lx->idx++;
                }
                if(lx->idx[0] == '>'){
                	lx->idx++;
					if(xdata->type == XML_TAG) xdata->sw_tag_open = 1;
					if(!parser(lx, xdata)){
						if(!lx->idx[0] && xdata->type == XML_TAG && xdata->sw_tag_open == 1 && !lx->sw_error)
							return error(lx, xdata->line, LT_ERROR_EOF);
						return 0;
					} else if(lx->idx[0] && lx->sw_error == 0){
						lx->idx++; xdata = 0;
						continue;
					} else
						return 1;
				}else if(lx->idx[0] == '/' && lx->idx[1] == '>'){
					if(xdata->type == XML_TAG) xdata->sw_tag_open = 0;
					tag_close = new_data(lx, XML_TAG_CLOSE, xdata->group_id);
					if(!tag_close) return error(lx, lx->line, LT_ERROR_MEM);
					tag_close->parent = xdata;
					lx->idx += 2;
					xdata = 0;
					continue;
                }
    			return error(lx, lx->line, "Unclosed tag");
            }
        }

        //tag content naar parent->val wegschrijven
        if(!parent && !xdata) return error(lx, lx->line, "Content not permitted");
		start = lx->idx; start_len = 0;
		while(lx->idx[0]){
			if(lx->idx[0] == ']' && lx->idx[1] == ']' && lx->idx[2] == '>')
				return error(lx, lx->line, "CDATA not permitted");

			if(lx->idx[0] == '<'){
				tagcontent = (parent ? parent->val : xdata->val);
				tagcontent = add_string(lx, tagcontent, start, start_len, 1);
                if(lx->sw_error) return 1;
				if(parent) parent->val = tagcontent; else xdata->val = tagcontent;
				break;
            }
            if(lx->idx[0] == '&'){
				refcnt = reference(lx, lx->idx, 0);
				if(!refcnt) return error(lx, lx->line, "Invalid entity reference");
				lx->idx += refcnt;
				start_len += refcnt;
				continue;
            }

			if(!ISVALID_VALUECHAR(lx->idx[0])) return error(lx, lx->line, "Invalid tag content");
			if(IS_EOL(lx->idx[0])) lx->line++;
			start_len++;
			lx->idx++;
		}
	}
	return 0;
}

//plays the memory tags sequentially
void lt_parse(LX_PARSER *lx){
	XML_DATA *elem = 0;
	size_t item = 0;

	if(!lx || (lx->OnBegin && !lx->OnBegin(lx))) return;
	while(item < lx->datacounter){
		elem = lx->data[item];
		if(!elem){ error(lx, lx->line, "Error while parsing, tag element is Null"); return; }
		switch(elem->type){
			case XML_TAG:
				if(lx->OnTag && lx->OnTag(lx, elem->name, (elem->val ? elem->val : (unsigned char*)""))) return;
				break;

			case XML_TAG_CLOSE:
				if(lx->OnTagEnd && lx->OnTagEnd(lx, elem->parent->name)) return;
				break;

			case XML_ATTRIB:
				if(lx->OnAttribute && lx->OnAttribute(lx, elem->parent->name, elem->name, elem->val)) return;
				break;
		}
		item++;
	}

	if(lx->OnFinnish) lx->OnFinnish(lx);

}

int error(LX_PARSER *lx, size_t line, char *errmsg){
	lx->sw_error = 1;
	if(lx->OnError) lx->OnError(lx, line, errmsg);
    return 0;
}

int lt_loadfilef(LX_PARSER *lx, FILE *f){
	struct stat st;
	size_t r = 0, add = 0;
		if(!lx) return 0;
		lt_reset(lx);
		lx->sw_error = 0;
		if(!f) return error(lx, 0, "Cannot read file.");
		if(0 != fstat(fileno(f), &st)){ fclose(f);  return error(lx, 0, "File size error"); }
		if(0 == st.st_size){ fclose(f);  return error(lx, 0, "File is empty"); }
		lx->bufsze = st.st_size;
		lx->buf = (unsigned char*)malloc(lx->bufsze + 1);
		if(!lx->buf){fclose(f);  return error(lx, 0, LT_ERROR_MEM); }

		do{
			add = (size_t)fread(lx->buf + r, sizeof(char), st.st_size, f);
			r += add;
		}while(add);
		fclose(f);

		lx->idx = lx->buf;
		lx->idx[r] = 0;
		if(encoding(lx)){
			lx->line = 1;
			parser(lx, 0);
		}

		free(lx->buf);
		return (lx->sw_error ? 0 : 1);
}

int lt_loadfile(LX_PARSER *lx, char *path){
	FILE *f;
	errno = 0;
	if(lx){
		f = fopen(path, "rb");
		if(f) return lt_loadfilef(lx, f);
	}
	return 0;
}
